import org.apache.commons.lang3.StringUtils

insight = new InsightTool(logger: logger, insightAccessToken: insightAccessToken, apiClientFactory: apiClientFactory)

/**
 * Velocity tool to get Insight object data from the Insight Cloud REST API.
 *
 * @see https://documentation.riada.io/display/INSCLOUD/REST+API
 */
class InsightTool {
	def logger

	/* Insight authentication parameter. */
	def insightAccessToken

	def apiClientFactory

	def insightClient

	def slurper = new groovy.json.JsonSlurper()

	/**
	 * Returns the Insight objects for the passed Asset field value (as Groovy objects slurped from <code>https://documentation.riada.io/display/INSCLOUD/REST+API+-+Load<code>)
	 * or the exception object if something went wrong.
	 */
	def getInsightObjects(def customFieldValue) {
		def insightObjectIds = customFieldValue.collect { StringUtils.substringAfterLast(it.originId, "_") }
		def insightObjects = []

		for (def id : insightObjectIds) {
			logger.log("Downloading Insight object with ID #${id}")
			def insightObject = callInsightApi("/rest/insight/1.0/object/${id}")
			if (!(insightObject instanceof Throwable)) {
				logger.log("Insight object found with ID #${id}")
				insightObjects.add(insightObject)
			} else {
				return insightObject
			}
		}

		return insightObjects
	}

	def getAttributeValues(def attribute) {
		attribute.objectAttributeValues.collect { it.displayValue }
	}

	/**
	 * Calls the Insight Cloud REST API.
	 */
	private def callInsightApi(url) {
		try {
			logger.log("Insight URL: ${url}")
			if (!insightClient) {
				insightClient = apiClientFactory.createInsightClient(insightAccessToken)
			}
			def json = insightClient.getJson(url)
			// logger.log("Insight JSON for ${url}: ${json}")

			def result = slurper.parseText(json)

			return result
		} catch (Throwable e) {
			logger.log("Failed to download Insight object data", e)
			return e
		}
	}
}
