import com.google.common.base.Predicate
import com.google.common.collect.Collections2
import org.joda.time.DateTimeConstants
import org.joda.time.Days
import org.joda.time.LocalDate

projectReport = new ProjectReportTool(componentAccessor, issues, startDateCustomFieldId, endDateCustomFieldId, periodLength, periodUnit)
filename = "${title} Status Report.pdf".toString() // "filename" is a standard Java String, not a GString

/**
 * Velocity tool to generate issue lists and Key Performance Indicators for project reports.
 * @see project-report-fo.vm as the actual template.
 */
public class ProjectReportTool {
	enum ProjectStatus { RED, YELLOW, GREEN }

	def componentAccessor

	def issues
	def startDateCustomFieldName
	def endDateCustomFieldName

	def now

	/* Reported period. */
	def startDate
	def endDate
	/* Next reported period. */
	def nextStartDate
	def nextEndDate
	/**
	 * IDs of the priorities that are considered "risks".
	 */
	def riskPriorityIds

	/**
	 * The priority with this ID (and all priorities ranked above this) is considered a "risk".
	 * Default setting: 2 is the ID of the default "High" priority, therefore any priority that is "High" or ranked above "High"
	 * is considered a risk.
	 */
	static final long LOWEST_RANKED_RISK_PRIORITY_ID = 2

	public ProjectReportTool(componentAccessor, issues, startDateCustomFieldId, endDateCustomFieldId, periodLength, periodUnit) {
		this.componentAccessor = componentAccessor
		this.issues = issues
		this.startDateCustomFieldName = customFieldIdToName(startDateCustomFieldId)
		this.endDateCustomFieldName = customFieldIdToName(endDateCustomFieldId)

		init(periodLength, periodUnit)
	}

	private String customFieldIdToName(customFieldId) {
		return ((customFieldId instanceof Integer) && !customFieldId.equals(0)) ? "customfield_${customFieldId}" : null
	}

	private def init(periodLength, periodUnit) {
		now = new LocalDate()
		if (periodUnit == 'month') {
			// current period: from the first day in this month until today (inclusive)
			startDate = now.dayOfMonth().withMinimumValue().toDate()
			endDate = now.toDate()

			// next period: from tomorrow until the last day in this month (inclusive)
			if (now.getMonthOfYear() == now.plusDays(1).getMonthOfYear()) {
				nextStartDate = now.plusDays(1).toDate()
				nextEndDate = now.dayOfMonth().withMaximumValue().toDate()
			}
		} else if (periodUnit == 'week') {
			int firstDayOfWeek = DateTimeConstants.MONDAY
			int lastDayOfWeek = DateTimeConstants.SUNDAY

			// current period: from the Monday this week until today (inclusive)
			startDate = now.withDayOfWeek(firstDayOfWeek).toDate()
			endDate = now.toDate()

			// next period: from tomorrow until the Sunday this week (inclusive)
			if (now.getWeekOfWeekyear() == now.plusDays(1).getWeekOfWeekyear()) {
				nextStartDate = now.plusDays(1).toDate()
				nextEndDate = now.withDayOfWeek(lastDayOfWeek).toDate()
			}
		} else { // last N days
			// current period: from N-day-earlier until today (inclusive)
			startDate = now.minusDays(periodLength - 1).toDate()
			endDate = now.toDate()

			// next period: from tomorrow until N-day-later (inclusive)
			nextStartDate = now.plusDays(1).toDate()
			nextEndDate = now.plusDays(periodLength).toDate()
		}

		riskPriorityIds = findRiskPriorityIds()
	}

	/**
	 * @see #LOWEST_RANKED_RISK_PRIORITY_ID
	 */
	def findRiskPriorityIds() {
		def riskPriorityIds = []
		for (priority in componentAccessor.constantsManager.priorities) {
			riskPriorityIds.add(priority.getId())
			if (priority.getId() == LOWEST_RANKED_RISK_PRIORITY_ID) {
				break
			}
		}

		return riskPriorityIds
	}

	ProjectStatus getOverallStatus() {
		def problemCount = getProblemCount()
		return (problemCount >= issues.size()/2) ? ProjectStatus.RED : ((problemCount == 0) ? ProjectStatus.GREEN : ProjectStatus.YELLOW)
	}

	int getTotalCount() {
		return issues.size()
	}

	int getTotalCompletedCount() {
		return Collections2.filter(issues, new IssueCompletedPredicate()).size()
	}

	int getCompletionPercentage() {
		return (issues.size() > 0) ? (100*Collections2.filter(issues, new IssueCompletedPredicate()).size()) / issues.size() : 0
	}

	int getProblemCount() {
		return getOverdueIssues().size() + getOverbudgetIssues().size() + getRiskIssues().size()
	}

	def getCompletedIssues() {
		return Collections2.filter(issues, new IssueCompletedBetweenPredicate(from: startDate, to: endDate))
	}

	def getPlannedIssues() {
		return (nextStartDate && nextEndDate) ? Collections2.filter(issues, new IssuePlannedBetweenPredicate(from: nextStartDate, to: nextEndDate)) : []
	}

	def getOverdueIssues() {
		return Collections2.filter(issues, new IssueOverduePredicate(now: now))
	}

	def getOverbudgetIssues() {
		return Collections2.filter(issues, new IssueOverbudgetPredicate())
	}

	def getRiskIssues() {
		return Collections2.filter(issues, new IssueRiskPredicate(riskPriorityIds: riskPriorityIds))
	}

	def getIssueStartDate(issue) {
		return (startDateCustomFieldName != null) ? issue.getCustomFieldValue(startDateCustomFieldName) : issue.created.clearTime()
	}

	def getIssueEndDate(issue) {
		return (endDateCustomFieldName != null) ? issue.getCustomFieldValue(endDateCustomFieldName) : issue.updated.clearTime()
	}

	def daysBetweenDateAndNow(date) {
		return (date != null) ? Days.daysBetween(new LocalDate(date), now).getDays() : -1
	}

	private class IssueCompletedPredicate implements Predicate {
		boolean apply(issue) {
			return (issue.resolutionDate != null)
		}
	}

	private class IssueCompletedBetweenPredicate implements Predicate {
		def from
		def to

		boolean apply(issue) {
			def resolutionDate = issue.resolutionDate?.clearTime()
			return (resolutionDate != null) && !resolutionDate.before(from) && !resolutionDate.after(to) // inclusive on both ends of the interval
		}
	}

	private class IssuePlannedBetweenPredicate implements Predicate {
		def from
		def to

		boolean apply(issue) {
			def start = getIssueStartDate(issue)
			def end = getIssueEndDate(issue)
			return (issue.resolutionDate == null) && (start != null) && (end != null) && !start.after(to) && !end.before(from) // inclusive on both ends of the interval
		}
	}

	private class IssueOverduePredicate implements Predicate {
		def now

		boolean apply(issue) {
			def end = getIssueEndDate(issue)
			return (issue.resolutionDate == null) && (end != null) && end.before(now.toDate()) // if "end date" is today, then it is not overdue
		}
	}

	private class IssueOverbudgetPredicate implements Predicate {
		boolean apply(issue) {
			return (issue.aggregateOriginalEstimate != null) && (issue.aggregateTimeSpent != null) && (issue.aggregateTimeSpent > issue.aggregateOriginalEstimate)
		}
	}

	private class IssueRiskPredicate implements Predicate {
		def riskPriorityIds

		boolean apply(issue) {
			return (issue.resolutionDate == null) && riskPriorityIds.contains(issue.priorityObject?.id)
		}
	}
}
