tempo = new TempoTool(logger: logger, componentAccessor: componentAccessor, tempoAccessToken: tempoAccessToken, apiClientFactory: apiClientFactory, dateFrom: tempoDateFrom, dateTo: tempoDateTo, issues: issues)

/**
 * Velocity tool to get Tempo data from the Tempo Cloud REST API.
 *
 * @see https://tempo-io.github.io/tempo-api-docs/
 */
class TempoTool {
	def logger

	/* Maximum number of issue keys to pass in the URL in a single REST API call. */
	def ISSUE_KEYS_PER_REQUEST = 50

	def componentAccessor

	/* Tempo authentication parameter. */
	def tempoAccessToken

	def apiClientFactory
	def tempoClient

	def dateFrom
	def dateTo

	def issues

	def preloadedAccounts = []
	def preloadedWorklogs = []
	def preloadedWorkAttributes = [:]

	def slurper = new groovy.json.JsonSlurper()

	/**
	 * Returns the account for the passed issue.
	 * It first checks the "Account" custom field, then it checks the "Account" type work attribute of the first worklog.
	 */
	def getAccount(def issue) {
		def account
		def accountCustomField = componentAccessor.customFieldManager.getCustomFieldObjects(issue).find { it.customFieldType.key == 'com.atlassian.plugins.atlassian-connect-plugin:io.tempo.jira__account' }
		def accountId = accountCustomField ? issue.getCustomFieldValue(accountCustomField)?.id : null
		account = accountId ? getAccountById(accountId) : null
		if (!account) {
			def accountAttribute = getAllWorkAttributes()?.values()?.find { it.type == 'ACCOUNT' }
			if (accountAttribute) {
				def worklogs = getWorklogs(issue.key)
				if (worklogs && !(worklogs instanceof Throwable)) {
					def accountKey = worklogs[0].attributes?.values?.find { it.key == accountAttribute.key }?.value
					account = accountKey ? getAccountByKey(accountKey) : null
				}
			}
		}
		account
	}

	/**
	 * Returns the worklogs for the passed issue as a <code>List</code> of <code>Maps</code>, or the exception object if something went wrong.
	 */
	def getWorklogs(def issueKey) {
		if (!preloadedWorklogs) {
			issues.collect { it.key }.collate(ISSUE_KEYS_PER_REQUEST).each { issueKeys ->
				if (!(preloadedWorklogs instanceof Throwable)) {
					logger.log("Downloading Tempo worklogs for ${issueKeys.join(', ')}")
					def result = callTempoApi("/worklogs?from=${dateFrom}&to=${dateTo}&issue=${issueKeys.join('&issue=')}", 'results')
					(!(result instanceof Throwable)) ? preloadedWorklogs.addAll(result.results) : (preloadedWorklogs = result)
				}
			}
		}
		!(preloadedWorklogs instanceof Throwable) ? preloadedWorklogs.findAll { it.issue.key == issueKey } : preloadedWorklogs
	}

	/**
	 * Returns comma-separated work attribute name-value pairs for the passed worklog.
	 */
	def getWorkAttributes(def worklog) {
		def workAttributes = getAllWorkAttributes()
		worklog.attributes.values.findAll { it.value }.collect {
			def attributeName = workAttributes ? workAttributes[it.key].name : it.key
			def attributeValue = it.value
			if (workAttributes && workAttributes[it.key].type == 'ACCOUNT') {
				attributeValue = getAccountByKey(it.value)?.name ?: it.value
			}
			"${attributeName}: ${attributeValue}"
		}
		.join(", ")
	}

	private def getAccountById(def accountId) {
		getAllAccounts()?.find { it.id == accountId }
	}

	private def getAccountByKey(def accountKey) {
		getAllAccounts()?.find { it.key == accountKey }
	}

	/**
	 * Returns all accounts configured in Tempo.
	 */
	private def getAllAccounts() {
		if (!preloadedAccounts) {
			def result = callTempoApi('/accounts')
			preloadedAccounts = !(result instanceof Throwable) ? result.results : result
		}
		!(preloadedAccounts instanceof Throwable) ? preloadedAccounts : null
	}

	/**
	 * Returns all work attributes configured in Tempo.
	 */
	private def getAllWorkAttributes() {
		if (!preloadedWorkAttributes) {
			def result = callTempoApi('/work-attributes')
			preloadedWorkAttributes = !(result instanceof Throwable) ? result.results.collectEntries {
				[(it.key): it]
			} : result
		}
		!(preloadedWorkAttributes instanceof Throwable) ? preloadedWorkAttributes : null
	}

	/**
	 * Calls the Tempo Cloud REST API and handles paging if needed.
	 * If the entityListFieldName parameter is passed, then that property will contain all the data of the same property of each page in the final result.
	 */
	private def callTempoApi(url, entityListFieldName = null) {
		try {
			logger.log("Tempo URL: ${url}")
			if (!tempoClient) {
				tempoClient = apiClientFactory.createTempoClient(tempoAccessToken)
			}
			def jsonPages = tempoClient.getJson(url)
			// logger.log("Tempo JSON for ${url}: ${jsonPages}")

			def result
			jsonPages.collect {
				def pageData = slurper.parseText(it)
				if (!result) {
					result = pageData
				} else {
					result[entityListFieldName].addAll(pageData[entityListFieldName])
				}
			}

			return result
		} catch (Throwable e) {
			logger.log('Failed to download Tempo data', e)
			return e
		}
	}
}