zephyr = new ZephyrTool(logger: logger, zephyrAccessKey: zephyrAccessKey, zephyrSecretKey: zephyrSecretKey, zephyrAccountId: zephyrAccountId, apiClientFactory: apiClientFactory, userManager: userManager)

/**
 * Velocity tool to get Zephyr test data from the Zephyr Cloud REST API.
 *
 * @see https://zephyrdocs.atlassian.net/wiki/spaces/ZFJCLOUD/pages/1925120024/REST+API
 */
class ZephyrTool {
	def logger

	/* Zephyr REST API authentication parameters. */
	def zephyrAccessKey
	def zephyrSecretKey
	def zephyrAccountId

	def apiClientFactory
	def userManager

	def zephyrClient

	def slurper = new groovy.json.JsonSlurper()

	/**
	 * Returns the Zephyr test steps for the passed issue (as Groovy objects slurped from <code>https://zfjcloud.docs.apiary.io/#reference/teststep/create-and-get-all-teststep/get-all-teststeps-v1<code>)
	 * or the exception object if something went wrong.
	 */
	def getTestSteps(def issue) {
		def issueId = issue.id
		def issueKey = issue.key
		def projectId = issue.projectObject.id
		logger.log("Downloading Zephyr test steps for issue #${issueId} [${issueKey}]")
		def testSteps = callZephyrApi("/public/rest/api/2.0/teststep/${issueId}?projectId=${projectId}", "testSteps")
		if (!(testSteps instanceof Throwable)) {
			logger.log("${testSteps.size()} Zephyr test steps found for issue #${issueId} [${issueKey}]")
		}
		return testSteps
	}

	/**
	 * Returns the Zephyr test executions for the passed issue (as Groovy objects slurped from <code>https://zfjcloud.docs.apiary.io/#reference/execution/get-list-of-executions/get-list-of-executions<code>)
	 * or the exception object if something went wrong.
	 */
	def getTestExecutions(def issue) {
		def issueId = issue.id
		def issueKey = issue.key
		def projectId = issue.projectObject.id
		logger.log("Downloading Zephyr test executions for issue #${issueId} [${issueKey}]")
		def testExecutions = callZephyrApi("/public/rest/api/1.0/executions?projectId=${projectId}&issueId=${issueId}", "executions")

		if (!(testExecutions instanceof Throwable)) {
			logger.log("${testExecutions.executions.size()} Zephyr test executions found for issue #${issueId} [${issueKey}]")

			// resolve account IDs to users
			def accountIds = testExecutions.executions.collect { it.execution.executedBy }.findAll { it != null }.unique()
			if (accountIds) {
				def users = userManager.getUsersByAccountId(accountIds)
				testExecutions.executions.each { execution ->
					def executedByUser = users.find { user ->
						user.accountId == execution.execution.executedByAccountId
					}
					execution.execution.executedBy = executedByUser?.displayName
				}
			}

			// sort
			testExecutions.executions = testExecutions.executions.sort { a, b -> a.execution.executedOn <=> b.execution.executedOn }.unique { a, b -> a.execution.id <=> b.execution.id } // FIXME Drop this when Zephyr fixes their API: We get the list of executions as an unordered list from Zephyr, therefore we sort it and remove duplicates which resulted by the buggy pagination.
		}

		return testExecutions
	}

	/**
	 * Returns the Zephyr test step results for the passed execution (as Groovy objects slurped from <code>https://zfjcloud.docs.apiary.io/#reference/stepresult/get-stepresults-by-execution/get-stepresults-by-execution<code>)
	 * or the exception object if something went wrong.
	 */
	def getTestStepResults(def executionId, def issue) {
		def issueId = issue.id
		logger.log("Downloading Zephyr test step results for execution #${executionId}")
		def testStepResults = callZephyrApi("/public/rest/api/1.0/stepresult/search?executionId=${executionId}&issueId=${issueId}&isOrdered=true", "stepResults")
		if (!(testStepResults instanceof Throwable)) {
			logger.log("${testStepResults.size()} Zephyr test step results found for execution #${executionId}")
		}
		return testStepResults
	}

	/**
	 * Calls the Zephyr Cloud REST API and handles paging if needed.
	 * If the entityListFieldName parameter is passed, then that property will contain all the data of the same property of each page in the final result.
	 */
	def callZephyrApi(url, entityListFieldName = null) {
		try {
			logger.log("Zephyr URL: ${url}")
			if (!zephyrClient) {
				zephyrClient = apiClientFactory.createZephyrClient(zephyrAccessKey, zephyrSecretKey, zephyrAccountId)
			}
			def jsonPages = zephyrClient.getJson(url)
			// logger.log("Zephyr JSON for ${url}: ${jsonPages}")

			def result
			jsonPages.collect {
				def pageData = slurper.parseText(it)
				if (!result) {
					result = pageData
				} else {
					result[entityListFieldName].addAll(pageData[entityListFieldName])
				}
			}

			return result
		} catch (Throwable e) {
			logger.log("Failed to download Zephyr test data", e)
			return e
		}
	}
}
